% The following code reproduces Figure 3 from the paper
% L. Demanet and P. Hand. Stable Optimizationless Recovery from Phaseless
% Linear Measurements.  J. Fourier Anal. Appl. 20, 199-221, 2014.
%
%
% This code solves PhaseLift and the corresponding feasibility problem:
%
% Let X be an n x n matrix.  Let A(X) be the result of a linear operator
% A(X)_i = <X, z_i z_i^t> for i = 1 ... m, where z_i are Gaussian.
%
% Phase lift is the convex program
%                 min tr(X) such that X>=0, A(X)=b
% where b = A(X0)+e, where X0 = x0 x0^t for a Gaussian x0 and ||e||_2 =
% epsilon.
%
% The feasibility problem is
%                 find X such that X>=0, A(X) = b.
%
% In panel (a), the feasibility problem is solved by a Douglas-Rachford
% method with epsilon = 0 and 0.1.  See section 6.2 of the paper for details.
% In panel (b), the feasibility problem is solved by a Nesterov Method with
% epsilon = 0 and 0.1.  See section 6.2 of the paper for details.
% In panel (c), PhaseLift is solved by a Nesterov method.  See section 6.2
% of the paper for details.
%


%% Define parameters
n = 30;
m = 180;
Z = randn(m,n);
x0 = randn(n,1); x0 = x0 / norm(x0);
b_exact = (Z*x0).^2;
epsilons = [0,1e-1];
X0 = x0*x0';
methods = {'Douglas-Rachford', 'Nesterov-feasibility', 'Nesterov'};
alpha = 2e-4;


%% Solve the semidefinite programs


n_it = 1000;  % Number of iterations
err = zeros(n_it, 2, length(methods));
% err is the collection of the relative errors indexed by iteration number,
% the value of epsilon, and the method type.

theta = 1;
for j = 1:2
    epsilon = epsilons(j)
    eta = randn(m,1);
    b = b_exact + epsilon*eta/norm(eta);
    P = @(X)scriptAstar(scriptinvAAstar(scriptA(X,Z),Z),Z);  % projection onto Range Astar
    Q = @(b)scriptAstar(scriptinvAAstar(b,Z),Z);    % Solve underdetermined least squares system A(X)=b
    Qb = Q(b);
    % Projection onto constraints A(X) = b is given by X = X - P(X) + Qb;
    
    G = @(X, lambda)scriptAstar(scriptA(X,Z)-b,Z) + lambda * eye(size(X));
    % G is the gradient of g(X) = 1/2 ||A(X) -b||^2 + lambda tr(X)
    
    for ll = 1:3
        method = methods{ll};
        
        X = zeros(n,n);
        
        if strcmp(method, 'Nesterov') || strcmp(method, 'Nesterov-feasibility')
            if strcmp(method, 'Nesterov')
                lambda = 0.0001;
            else
                lambda = 0;
            end
            
            X_old = X;
            Y_old = X;
            theta_old = 1;
            for k= 1:n_it
                X_new = Ppos(Y_old - alpha*G(Y_old, lambda));
                theta_new = 2 / (1 + sqrt(1 + 4 / (theta_old^2)));
                beta = theta_new * (1/theta_old - 1);
                Y_new = X_new + beta * (X_new - X_old);
                
                X_old = X_new;
                Y_old = Y_new;
                theta_old = theta_new;
                
                err(k,j, ll) = norm(X_new-X0,'fro')/norm(X0,'fro');
            end
        elseif strcmp(method, 'Douglas-Rachford')
            Y = X;
            for k = 1:n_it
                % proj onto constraints AX = b is given by X = X - P(X) + Qb;
                Y = X - P(2*X-Y) + Qb;
                
                % proj onto sdp cone
                X = Ppos(Y);
                err(k,j, ll) = norm(X-X0,'fro')/norm(X0,'fro');
            end
            
        end
        
    end
end

%% Make plots
set(0, 'defaulttextinterpreter', 'none')
figure(1);close;figure(1)
set(gcf, 'Position', [360, 345, 900, 300]);

for ll=1:length(methods)
    method = methods{ll}
    sp{ll} = subplot(1,length(methods),ll)
    a = gca;
    set(a, 'FontSize', 18)
    set(a, 'Box', 'off')
    
    
    h = semilogy(1:n_it,err(:,2, ll), 'k', 'LineWidth', 1)
    set(h, 'color', [0.5, 0.5, 0.5])
    hold on;
    semilogy(1:n_it,err(:,1, ll), 'k', 'LineWidth', 2)
    axis([0, 1000, 1e-16, 1])
    set(gca,'XTick',[0,500,1000])
    set(gca,'YTick',10.^(-[15:-5:0]))
    
    if strcmp(method, 'Nesterov')
        title_str = 'Nesterov - PhaseLift';
    elseif strcmp(method, 'Nesterov-feasibility')
        title_str = 'Nesterov - Feasibility';
    elseif strcmp(method, 'Douglas-Rachford')
        title_str = 'Douglas-Rachford - Feasibility';
    end
    
    text(500,100, title_str, 'VerticalAlignment', 'middle', 'HorizontalAlignment', 'center', 'FontSize', 18)
    
    
    if ll==3
        lh = legend('Noisy: $\eps = 10^{-1}$','Noiseless: $\eps = 0$','Location', [0.65, 0.2, 0.3539, 0.2254])
    end
    
    xlabel('Iterations')
    
    if ll==1
        ylabel('Relative Recovery Error');
    end
    
    legend boxoff
end
