% The following code solves a single semidefinite feasibility problem (2) in
% L. Demanet and P. Hand. Stable Optimizationless Recovery from Phaseless
% Linear Measurements.  J. Fourier Anal. Appl. 20, 199-221, 2014.
%
%
% Let X be an n x n matrix.  Let A(X) be the result of a linear operator
% A(X)_i = <X, z_i z_i^t> for i = 1 ... m, where z_i are Gaussian.
%
%
% The feasibility problem (2) is
%                 find X such that X>=0, A(X) = b.
%
% where b = A(X0) + e, where ||e||=epsilon and X0 = x0*x0' and x0 is a
% Gaussian normalized to have unit l2 norm.  
%
% This code solves the semidefinite feasibility problem 
%                 find X >=0 such that A(X) = b
%
% The feasibility problem is solved by a Douglas-Rachford method. See
% section 6.2 of the paper for details.

n = 30;           % Ambient dimension
m = 140;          % Number of phaseless random gaussian measurements
epsilon = 1e-1;   % L2 norm of error in measurement vector
n_it = 1e3;       % Number of iterations


% Set up problem
Z = randn(m,n);                        % Columns are random measurement vectors
x0 = randn(n,1); x0 = x0 / norm(x0);   % Unit norm true solution
X0 = x0*x0';                           % The true matrix to recover
b_exact = (Z*x0).^2;                   % Exact measurement vector
eta = randn(m,1);
b = b_exact + epsilon*eta/norm(eta);   % Noisy measurement vector

if m >= n*(n+1)/2
    % In this regime, the problem is critically determined or
    % overdetermined.  Solve by least squares.    
    X = solve_scriptAx_b(b, Z);
else
    P = @(X)scriptAstar(scriptinvAAstar(scriptA(X,Z),Z),Z); % projection onto range Astar
    Q = @(b)scriptAstar(scriptinvAAstar(b,Z),Z);    % underdetermined least squares solution to A(X)=b
    Qb = Q(b);
    resids = zeros(n_it, 1);
    
    X = zeros(n,n);
    Y = X;
    % Execute Douglas-Rachford method
    for k = 1:n_it
        % Projection onto constraints A(X) = b is given by X = X - P(X) + Qb;
        Y = X - P(2*X-Y) + Qb;
        % Projection onto postive semidefinite cone
        X = Ppos(Y);        
    end
end

% Extract the largest eigenvector of X
[V,D] = eig((X+X')/2);
index_max = find(diag(D)==max(abs(diag(D))));
index_max = index_max(1);
x = V(:, index_max);
err = min( norm(x - x0), norm(x+x0)); % x may be off by a global minus sign

display(sprintf('Recovering the %d-vector from %d phaseless measurements with error parameter %d\n', n, m, epsilon))
display('The unknown vector is')
x0'
display('The recovered vector iss')
x'
display(sprintf('The relative error is %d', err))

